import unittest

import numpy as np
import matplotlib.pyplot as pl
import networkx as nx

from netwulf.tools import bind_properties_to_network, get_filtered_network, draw_netwulf
from netwulf import visualize

def _get_test_network():
    G = nx.Graph()
    G.add_nodes_from(range(2))
    G.add_nodes_from("ab")
    G.add_edges_from([("a","b")])
    G.add_edges_from([("a",0)])

    return G

def _get_test_config():
    return {'node_size':2,'link_width':2}

def _drastically_round_positions(props,nprec=-1):

    for i, n in enumerate(props['nodes']):
        props['nodes'][i]['x'] = np.around(props['nodes'][i]['x'],nprec)
        props['nodes'][i]['y'] = np.around(props['nodes'][i]['y'],nprec)
        props['nodes'][i]['x_canvas'] = np.around(props['nodes'][i]['x_canvas'],nprec)
        props['nodes'][i]['y_canvas'] = np.around(props['nodes'][i]['y_canvas'],nprec)

def _assert_positions_within_one_percent(props1,props2):

    for i, n in enumerate(props1['nodes']):
        assert(np.isclose(props1['nodes'][i]['x'], props2['nodes'][i]['x'],rtol=1e-2))
        assert(np.isclose(props1['nodes'][i]['y'], props2['nodes'][i]['y'],rtol=1e-2))
        assert(np.isclose(props1['nodes'][i]['x_canvas'], props2['nodes'][i]['x_canvas'],rtol=1e-2))
        assert(np.isclose(props1['nodes'][i]['y_canvas'], props2['nodes'][i]['y_canvas'],rtol=1e-2))


class Test(unittest.TestCase):

    maxDiff = None

    def test_posting(self):
        """Test whether results are sucessfully posted to Python."""
        G = _get_test_network()
        visualize(G,is_test=True,config=_get_test_config())

    def test_reproducibility(self):
        """Test whether a restarted network visualization with binded positions results in the same visualization."""
        G = _get_test_network()
        props, config = visualize(G,config=_get_test_config(),is_test=True)        

        bind_properties_to_network(G, props)
        newprops, newconfig = visualize(G, config=config,is_test=True)

        # test if node positions are close within 1% and subsequently 
        # round all positional values to the 2nd positon (e.g. 451 => 450)
        _assert_positions_within_one_percent(props,newprops)
        _drastically_round_positions(props)
        _drastically_round_positions(newprops)

        # the second visualization should have frozen the nodes
        assert(newconfig['freeze_nodes'])

        # change it so the dictionary-equal-assertion doesnt fail
        newconfig['freeze_nodes'] = False

        self.assertDictEqual(props, newprops)
        self.assertDictEqual(config, newconfig)
    
    def test_filtering(self):
        """Test whether filtering works the way it should."""

        G = _get_test_network()
        weights = [10,100]
        for e, (u, v) in enumerate(G.edges()):
            G[u][v]['foo'] = weights[e]
            G[u][v]['bar'] = weights[(e+1)%2]

        grp = {u: 'AB'[i%2]  for i, u in enumerate(G.nodes()) }

        new_G = get_filtered_network(G,edge_weight_key='foo')
        visualize(new_G,is_test=True,config=_get_test_config())

        nx.set_node_attributes(G, grp, 'wum')

        new_G = get_filtered_network(G,edge_weight_key='bar',node_group_key='wum')
        visualize(new_G,is_test=True,config=_get_test_config())

    def test_matplotlib(self):
        """Test how the produced figure looks"""
        stylized_network = {   'linkAlpha': 0.2,
            'linkColor': '#758000',
            'links': [   {'source': 0, 'target': 1, 'weight': 1, 'width': 40},
                         {'source': 0, 'target': 3, 'weight': 1, 'width': 40},
                         {'source': 1, 'target': 2, 'weight': 1, 'width': 40},
                         {'source': 1, 'target': 4, 'weight': 1, 'width': 40}],
            'nodeStrokeColor': '#ffffff',
            'nodeStrokeWidth': 4,
            'nodes': [   {   'color': '#79aa00',
                             'id': 0,
                             'radius': 28.577380332470412,
                             'x': 420.4774227636891,
                             'x_canvas': 444.34195934582385,
                             'y': 402.92277557839725,
                             'y_canvas': 321.45942904878075},
                         {   'color': '#79aa00',
                             'id': 1,
                             'radius': 35,
                             'x': 419.94297017162165,
                             'x_canvas': 440.6007912013515,
                             'y': 426.56813829452074,
                             'y_canvas': 486.9769680616455},
                         {   'color': '#79aa00',
                             'id': 2,
                             'radius': 20.207259421636905,
                             'x': 405.5333864027982,
                             'x_canvas': 339.73370481958773,
                             'y': 411.01250794520007,
                             'y_canvas': 378.08755561640055},
                         {   'color': '#79aa00',
                             'id': 3,
                             'radius': 20.207259421636905,
                             'x': 434.1529311105535,
                             'x_canvas': 540.0705177738741,
                             'y': 413.3758620827409,
                             'y_canvas': 394.6310345791867},
                         {   'color': '#79aa00',
                             'id': 4,
                             'radius': 20.207259421636905,
                             'x': 403.23992572436083,
                             'x_canvas': 323.67948007052564,
                             'y': 429.81017078651104,
                             'y_canvas': 509.67119550557754}],
            'xlim': [0, 833],
            'ylim': [0, 833]
            }


        draw_netwulf(stylized_network)
        pl.show(block=False)
        pl.pause(5)
        pl.close()


    def test_config_adaption(self):
        """Test whether config values are properly adapted."""
        config = {
            # Input/output
            'zoom':4,
            # Physics
            'node_charge': -70,
            'node_gravity': 0.5,
            'link_distance': 15,
            'link_distance_variation': 2,
            'node_collision': True,
            'wiggle_nodes': True,
            'freeze_nodes': False,
            # Nodes
            'node_fill_color': '#79aa00',
            'node_stroke_color': '#ffffff',
            'node_label_color': '#888888',
            'display_node_labels': True,
            'scale_node_size_by_strength': True,
            'node_size': 5,
            'node_stroke_width': 1,
            'node_size_variation': 0.7,
            # Links
            'link_color': '#758000',
            'link_width': 10,
            'link_alpha': 0.2,
            'link_width_variation': 0.7,
            # Thresholding
            'display_singleton_nodes': False,
            'min_link_weight_percentile': 0.1,
            'max_link_weight_percentile': 0.9,
        }

        G = _get_test_network()
        _, newconfig = visualize(G,config=config,is_test=True)

        self.assertDictEqual(config, newconfig)



if __name__ == "__main__":


    T = Test()

    #T.test_matplotlib()
    T.test_config_adaption()



from setuptools import setup, Extension
import setuptools
import os, sys

# get __version__, __author__, and __email__
exec(open("./netwulf/metadata.py").read())

setup(
    name = 'netwulf',
    version = __version__,
    author = __author__,
    author_email = __email__,
    url = 'https://github.com/benmaier/netwulf',
    license = __license__,
    description = "Interactively visualize networks with Ulf Aslak's d3-tool from Python.",
    long_description = '',
    packages = setuptools.find_packages(),
    python_requires='~=3.5', 
    install_requires = [
                'networkx>=2.0',
                'numpy>=0.14',
                'matplotlib>=3.0',
                'simplejson>=3.0',
            ],
    tests_require=['pytest', 'pytest-cov'],
    setup_requires=['pytest-runner'],
    classifiers=['License :: OSI Approved :: MIT License',
                'Programming Language :: Python :: 3.5',
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: 3.7'
               ],
    project_urls={
        'Documentation': 'https://netwulf.rtfd.io',
        'Contributing Statement': 'https://github.com/benmaier/netwulf/blob/master/CONTRIBUTING.md',
        'Bug Reports': 'https://github.com/benmaier/netwulf/issues',
        'Source': 'https://github.com/benmaier/netwulf/',
        'PyPI': 'https://pypi.org/project/netwulf/',
    },
    include_package_data = True,
    zip_safe = False,
)



"""
Some useful things to tweak and reproduce the visualizations.
"""

import numpy as np
import networkx as nx

import matplotlib as mpl
import matplotlib.pyplot as pl
from matplotlib.collections import LineCollection, EllipseCollection


def bind_properties_to_network(network,
                               network_properties,
                               bind_node_positions=True,
                               bind_node_color=True,
                               bind_node_radius=True,
                               bind_node_stroke_color=True,
                               bind_node_stroke_width=True,
                               bind_link_width=True,
                               bind_link_color=True,
                               bind_link_alpha=True):
    """
    Binds calculated positional values to the network as node attributes `x` and `y`.

    Parameters
    ----------
    network : networkx.Graph or something alike
        The network object to which the position should be bound
    network_properties : dict
        The network properties which are returned from the
        interactive visualization.
    bind_node_positions : bool (default: True)
    bind_node_color : bool (default: True)
    bind_node_radius : bool (default: True)
    bind_node_stroke_color : bool (default: True)
    bind_node_stroke_width : bool (default: True)
    bind_link_width : bool (default: True)
    bind_link_color : bool (default: True)
    bind_link_alpha : bool (default: True)

    Example
    -------
        >>> props, _ = visualize(G)
        >>> bind_properties_to_network(G, props)
    """
    # Add individial node attributes
    if bind_node_positions:
        x = { node['id']: node['x'] for node in network_properties['nodes'] }
        y = { node['id']: node['y'] for node in network_properties['nodes'] }
        nx.set_node_attributes(network, x, 'x')
        nx.set_node_attributes(network, y, 'y')
        network.graph['rescale'] = False
    if bind_node_color:
        color = { node['id']: node['color'] for node in network_properties['nodes'] }
        nx.set_node_attributes(network, color, 'color')
    if bind_node_radius:
        radius = { node['id']: node['radius'] for node in network_properties['nodes'] }
        nx.set_node_attributes(network, radius, 'radius')

    # Add individual link attributes
    if bind_link_width:
        width = { (link['source'], link['target']): link['width'] for link in network_properties['links'] }
        nx.set_edge_attributes(network, width, 'width')

    # Add global style properties
    if bind_node_stroke_color:
        network.graph['nodeStrokeColor'] = network_properties['nodeStrokeColor']
    if bind_node_stroke_width:
        network.graph['nodeStrokeWidth'] = network_properties['nodeStrokeWidth']
    if bind_link_color:
        network.graph['linkColor'] = network_properties['linkColor']
    if bind_link_alpha:
        network.graph['linkAlpha'] = network_properties['linkAlpha']

def get_filtered_network(network,edge_weight_key=None,node_group_key=None):
    """
    Get a copy of a network where the edge attribute ``'weight'`` is
    set to the attribute given by the keyword ``edge_weight_key`` and the
    nodes are regrouped according to their node attribute provided by
    ``node_group_key``. 

    Parameters
    ----------
    network : networkx.Graph or alike
        The network object which is about to be filtered
    edge_weight_key : str, default : None
        If provided, set the edge weight to the edge attribute
        given by ``edge_weight_key`` and delete all other edge attributes
    node_group_key : str, default : None
        If provided, set the node ``'group'`` attribute according to a
        new grouping provided by the node attribute ``node_group_key``.

    Returns
    -------
    G : networkx.Graph or alike
        A filtered copy of the original network.
    """

    G = network.copy()

    if edge_weight_key is not None:
        for u, v, d in G.edges(data=True):
            keep_value = d[edge_weight_key]
            d.clear()
            G[u][v]['weight'] = keep_value

    if node_group_key is not None:
        groups = { node[1][node_group_key] for node in network.nodes(data=True) }
        groups_enum = {v: k for k,v in enumerate(groups)}
        for u in network.nodes():
            grp = G.node[u].pop(node_group_key)
            keep_value = groups_enum[grp]
            G.node[u]['group'] = keep_value

    return G

def draw_netwulf(network_properties, fig=None, ax=None, figsize=None):
    """
    Redraw the visualization using matplotlib. Creates
    figure and axes if None provided.
    In order to add labels, do for instance

    .. code:: python

        ax.text(
                network_properties['nodes'][0]['x'],
                network_properties['nodes'][0]['y'],
                network_properties['nodes'][0]['id']
               )

    Parameters
    ----------
    network_properties : dict
        The network properties which are returned from the
        interactive visualization.
    fig : matplotlib.Figure, default : None
        The figure in which to draw
    ax : matplotlib.Axes, default : None
        The Axes in which to draw
    figsize : float, default : None
        the size of the figure in inches (sidelength of a square)
        if None, will be taken as the minimum of the values in 
        ``matplotlib.rcParams['figure.figsize']``.
    
    Returns
    -------
    fig : matplotlib.Figure, default : None
        Resulting figure
    ax : matplotlib.Axes, default : None
        Resulting axes
    """

    # if no figure given, create a square one
    if ax is None or fig is None:
        if figsize is None:
            size = min(mpl.rcParams['figure.figsize'])
        else:
            size = figsize

        fig = pl.figure(figsize=(size,size))
        ax = fig.add_axes([0, 0, 1, 1])
        # Customize the axis
        # remove top and right spines
        ax.spines['right'].set_color('none')
        ax.spines['left'].set_color('none')
        ax.spines['top'].set_color('none')
        ax.spines['bottom'].set_color('none')
        # turn off ticks
        ax.xaxis.set_ticks_position('none')
        ax.yaxis.set_ticks_position('none')
        ax.xaxis.set_ticklabels([])
        ax.yaxis.set_ticklabels([])


    # for conversion of inches to points
    # (important for markersize and linewidths)
    dpi = fig.dpi

    # set everything square and get the axis size in points
    ax.axis('square')
    ax.axis('off')
    ax.margins(0)
    ax.set_xlim(network_properties['xlim'])
    ax.set_ylim(network_properties['ylim'])
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    axwidth, axheight = bbox.width*dpi, bbox.height*dpi



    # filter out node positions for links
    width = network_properties['xlim'][1] - network_properties['xlim'][0]
    height = network_properties['ylim'][1] - network_properties['ylim'][0]
    pos = { node['id']: np.array([node['x_canvas'], height - node['y_canvas']]) for node in network_properties['nodes'] }

    lines = []
    linewidths = []
    for link in network_properties['links']:
        u, v = link['source'], link['target']
        lines.append([ 
            [pos[u][0], pos[v][0]], 
            [pos[u][1], pos[v][1]]
                     ])
        linewidths.append(link['width']/width*axwidth)

    # collapse to line segments
    lines = [list(zip(x, y)) for x, y in lines]

    # plot Lines
    alpha = network_properties['linkAlpha']
    color = network_properties['linkColor']
    ax.add_collection(LineCollection(lines, 
                                     color=color,
                                     alpha=alpha, 
                                     linewidths=linewidths,
                                     zorder=-1
                                     ))

    # compute node positions and properties
    XY = []
    size = []
    node_colors = []

    for node in network_properties['nodes']:
        XY.append([node['x_canvas'], height - node['y_canvas']])
        # size has to be given in points*2
        size.append( 2*node['radius'] )
        node_colors.append(node['color'])

    XY = np.array(XY)
    size = np.array(size)
    circles = EllipseCollection(size,size,np.zeros_like(size),
                                offsets=XY,
                                units='x',
                                transOffset=ax.transData,
                                facecolors=node_colors,
                                linewidths=network_properties['nodeStrokeWidth']/width*axwidth,
                                edgecolors=network_properties['nodeStrokeColor'],
                                )
    ax.add_collection(circles)

    return fig, ax


